# todo: don't hard code this
version = \
'limes version 0.4.dev3'
help = '\n\n\
%s\n\
\n\
To run limes, one of the following is required:\n\
\n\
-h, --help\n\
    use with limes [command] -h to get sub options, example: limes login -h\n\
\n\
login\n\
    login to eLab using existing credentials\n\
    you will be prompted for a username and password\n\
\n\
get\n\
    *todo\n\
\n\
add\n\
    *todo\n\
' % version